from __future__ import division, print_function

from contextlib import contextmanager

import numpy as np


@contextmanager
def setattr_context(obj, **kwargs):
    """
    Context manager to temporarily change the values of object attributes
    while executing a function.

    Example
    -------
    >>> class Foo: pass
    >>> f = Foo(); f.attr = 'hello'
    >>> with setattr_context(f, attr='goodbye'):
    ...     print(f.attr)
    goodbye
    >>> print(f.attr)
    hello
    """
    old_kwargs = dict([(key, getattr(obj, key)) for key in kwargs])
    [setattr(obj, key, val) for key, val in kwargs.items()]
    try:
        yield
    finally:
        [setattr(obj, key, val) for key, val in old_kwargs.items()]


def iterable(obj):
    """Utility to check if object is iterable"""
    try:
        iter(obj)
    except:
        return False
    else:
        return True


def validate_inputs(*arrays, **kwargs):
    """Validate input arrays

    This checks that
    - Arrays are mutually broadcastable
    - Broadcasted arrays are one-dimensional

    Optionally, arrays are sorted according to the ``sort_by`` argument.

    Parameters
    ----------
    *args : ndarrays
        All non-keyword arguments are arrays which will be validated
    sort_by : array
        If specified, sort all inputs by the order given in this array.
    """
    arrays = np.broadcast_arrays(*arrays)
    sort_by = kwargs.pop('sort_by', None)

    if kwargs:
        raise ValueError("unrecognized arguments: {0}".format(kwargs.keys()))

    if arrays[0].ndim != 1:
        raise ValueError("Input arrays should be one-dimensional.")

    if sort_by is not None:
        isort = np.argsort(sort_by)
        if isort.shape != arrays[0].shape:
            raise ValueError("sort shape must equal array shape.")
        arrays = tuple([a[isort] for a in arrays])
    return arrays


def windowed_sum_slow(arrays, span, t=None, indices=None, tpowers=0,
                      period=None, subtract_mid=False):
    """Compute the windowed sum of the given arrays.

    This is a slow function, used primarily for testing and validation
    of the faster version of ``windowed_sum()``

    Parameters
    ----------
    arrays : tuple of arrays
        arrays to window
    span : int or array of ints
        The span to use for the sum at each point. If array is provided,
        it must be broadcastable with ``indices``
    indices : array
        the indices of the center of the desired windows. If ``None``,
        the indices are assumed to be ``range(len(arrays[0]))`` though
        these are not actually instantiated.
    t : array (optional)
        Times associated with the arrays
    tpowers : list (optional)
        Powers of t for each array sum
    period : float (optional)
        Period to use, if times are periodic. If supplied, input times
        must be arranged such that (t % period) is sorted!
    subtract_mid : boolean
        If true, then subtract the middle value from each sum

    Returns
    -------
    arrays : tuple of ndarrays
        arrays containing the windowed sum of each input array
    """
    span = np.asarray(span, dtype=int)
    if not np.all(span > 0):
        raise ValueError("span values must be positive")
    
    arrays = tuple(map(np.asarray, arrays))
    N = arrays[0].size
    if not all(a.shape == (N,) for a in arrays):
        raise ValueError("sizes of provided arrays must match")
    
    t_input = t
    if t is not None:
        t = np.asarray(t)
        if t.shape != (N,):
            raise ValueError("shape of t must match shape of arrays")
    else:
        t = np.ones(N)
    
    tpowers = tpowers + np.zeros(len(arrays))
    if len(tpowers) != len(arrays):
        raise ValueError("tpowers must be broadcastable with number of arrays")

    if period:
        if t_input is None:
            raise ValueError("periodic requires t to be provided")
        t = t % period
    
    if indices is None:
        indices = np.arange(N)
    spans, indices = np.broadcast_arrays(span, indices)
    
    results = []
    for tpower, array in zip(tpowers, arrays):
        if period:
            result = [sum(array[j % N]
                          * (t[j % N] + (j // N) * period) ** tpower
                          for j in range(i - s // 2,
                                         i - s // 2 + s)
                          if not (subtract_mid and j == i))
                      for i, s in np.broadcast(indices, spans)]
        else:
            result = [sum(array[j] * t[j] ** tpower
                          for j in range(max(0, i - s // 2),
                                         min(N, i - s // 2 + s))
                          if not (subtract_mid and j == i))
                      for i, s in np.broadcast(indices, spans)]
        results.append(np.asarray(result))
        
    return tuple(results)


def windowed_sum(arrays, span, t=None, indices=None, tpowers=0,
                 period=None, subtract_mid=False):
    """Compute the windowed sum of the given arrays.

    Parameters
    ----------
    arrays : tuple of arrays
        arrays to window
    span : int or array of ints
        The span to use for the sum at each point. If array is provided,
        it must be broadcastable with ``indices``
    indices : array
        the indices of the center of the desired windows. If ``None``,
        the indices are assumed to be ``range(len(arrays[0]))`` though
        these are not actually instantiated.
    t : array (optional)
        Times associated with the arrays
    tpowers : list (optional)
        Powers of t for each array sum
    period : float (optional)
        Period to use, if times are periodic. If supplied, input times
        must be arranged such that (t % period) is sorted!
    subtract_mid : boolean
        If true, then subtract the middle value from each sum

    Returns
    -------
    arrays : tuple of ndarrays
        arrays containing the windowed sum of each input array
    """
    span = np.asarray(span, dtype=int)
    if not np.all(span > 0):
        raise ValueError("span values must be positive")
    
    arrays = tuple(map(np.asarray, arrays))
    N = arrays[0].size
    if not all(a.shape == (N,) for a in arrays):
        raise ValueError("sizes of provided arrays must match")
    
    t_input = t
    if t is not None:
        t = np.asarray(t)
        if t.shape != (N,):
            raise ValueError("shape of t must match shape of arrays "
                             "t -> {0} arr -> {1}".format(t.shape,
                                                          arrays[0].shape))
    else:
        # XXX: special-case no t?
        t = np.ones(N)
    
    tpowers = np.asarray(tpowers) + np.zeros(len(arrays))

    if indices is not None:
        span, indices = np.broadcast_arrays(span, indices)

    # For the periodic case, re-call the function with padded arrays
    if period:
        if t_input is None:
            raise ValueError("periodic requires t to be provided")
        t = t % period

        t, arrays, sl = _pad_arrays(t, arrays, indices, span, period)
        if len(t) > N:
            # arrays are padded. Recursively call windowed_sum() and return.
            if span.ndim == 0 and indices is None:
                # fixed-span/no index case is done faster this way
                arrs = windowed_sum(arrays, span, t=t, indices=indices,
                                    tpowers=tpowers, period=None,
                                    subtract_mid=subtract_mid)
                return tuple([a[sl] for a in arrs])
            else:
                # this works for variable span and general indices
                if indices is None:
                    indices = np.arange(N)
                indices = indices + sl.start
                return windowed_sum(arrays, span, t=t, indices=indices,
                                    tpowers=tpowers, period=None,
                                    subtract_mid=subtract_mid)
        else:
            # No padding needed! We can carry-on as if it's a non-periodic case
            period = None

    # The rest of the algorithm now proceeds without reference to the period
    # just as a sanity check...
    assert period is None

    if span.ndim == 0:
        # fixed-span case. Because of the checks & manipulations above
        # we know here that indices=None
        assert indices is None
        window = np.ones(span)

        def convolve_same(a, window):
            if len(window) <= len(a):
                res = np.convolve(a, window, mode='same')
            else:
                res = np.convolve(a, window, mode='full')
                start = (len(window) - 1) // 2
                res = res[start:start + len(a)]
            return res
        results = [convolve_same(a * t ** tp, window)
                   for a, tp in zip(arrays, tpowers)]
        indices = slice(None)

    else:
        # variable-span case. Use reduceat() in a clever way for speed.
        if indices is None:
            indices = np.arange(len(span))

        # we checked this above, but just as a sanity check assert it here...
        assert span.shape == indices.shape

        mins = np.asarray(indices) - span // 2
        results = []
        for a, tp in zip(arrays, tpowers):
            ranges = np.vstack([np.maximum(0, mins),
                                np.minimum(len(a), mins+span)]).ravel('F')
            results.append(np.add.reduceat(np.append(a * t ** tp, 0),
                                           ranges)[::2])

    # Subtract the midpoint if required: this is used in cross-validation
    if subtract_mid:
        results = [r - a[indices] * t[indices] ** tp
                   for r, a, tp in zip(results, arrays, tpowers)]

    return tuple(results)


def _pad_arrays(t, arrays, indices, span, period):
    """Internal routine to pad arrays for periodic models."""
    N = len(t)

    if indices is None:
        indices = np.arange(N)
    pad_left = max(0, 0 - np.min(indices - span // 2))
    pad_right = max(0, np.max(indices + span - span // 2) - (N - 1))

    if pad_left + pad_right > 0:
        Nright, pad_right = divmod(pad_right, N)
        Nleft, pad_left = divmod(pad_left, N)
        t = np.concatenate([t[N - pad_left:] - (Nleft + 1) * period]
                           + [t + i * period
                              for i in range(-Nleft, Nright + 1)]
                           + [t[:pad_right] + (Nright + 1) * period])
        arrays = [np.concatenate([a[N - pad_left:]]
                                 + (Nleft + Nright + 1) * [a]
                                 + [a[:pad_right]])
                  for a in arrays]
        pad_left = pad_left % N
        Nright = pad_right / N
        pad_right = pad_right % N

        return (t, arrays, slice(pad_left + Nleft * N,
                                 pad_left + (Nleft + 1) * N))
    else:
        return (t, arrays, slice(None))


def _prep_smooth(t, y, dy, span, t_out, span_out, period):
    """Private function to prepare & check variables for smooth utilities"""

    # If period is provided, sort by phases. Otherwise sort by t
    if period:
        t = t % period
        if t_out is not None:
            t_out = t_out % period

    t, y, dy = validate_inputs(t, y, dy, sort_by=t)

    if span_out is not None:
        if t_out is None:
            raise ValueError("Must specify t_out when span_out is given")
        if span is not None:
            raise ValueError("Must specify only one of span, span_out")
        span, t_out = np.broadcast_arrays(span_out, t_out)
        indices = np.searchsorted(t, t_out)
    elif span is None:
        raise ValueError("Must specify either span_out or span")
    else:
        indices = None

    return t, y, dy, span, t_out, span_out, indices


def moving_average_smooth(t, y, dy, span=None, cv=True,
                          t_out=None, span_out=None, period=None):
    """Perform a moving-average smooth of the data

    Parameters
    ----------
    t, y, dy : array_like
        time, value, and error in value of the input data
    span : array_like
        the integer spans of the data
    cv : boolean (default=True)
        if True, treat the problem as a cross-validation, i.e. don't use
        each point in the evaluation of its own smoothing.
    t_out : array_like (optional)
        the output times for the moving averages
    span_out : array_like (optional)
        the spans associated with the output times t_out
    period : float
        if provided, then consider the inputs periodic with the given period

    Returns
    -------
    y_smooth : array_like
        smoothed y values at each time t (or t_out)
    """
    prep = _prep_smooth(t, y, dy, span, t_out, span_out, period)
    t, y, dy, span, t_out, span_out, indices = prep

    w = dy ** -2
    w, yw = windowed_sum([w, y * w], t=t, span=span, subtract_mid=cv,
                         indices=indices, period=period)

    if t_out is None or span_out is not None:
        return yw / w
    else:
        i = np.minimum(len(t) - 1, np.searchsorted(t, t_out))
        return yw[i] / w[i]


def linear_smooth(t, y, dy, span=None, cv=True,
                  t_out=None, span_out=None, period=None):
    """Perform a linear smooth of the data

    Parameters
    ----------
    t, y, dy : array_like
        time, value, and error in value of the input data
    span : array_like
        the integer spans of the data
    cv : boolean (default=True)
        if True, treat the problem as a cross-validation, i.e. don't use
        each point in the evaluation of its own smoothing.
    t_out : array_like (optional)
        the output times for the moving averages
    span_out : array_like (optional)
        the spans associated with the output times t_out
    period : float
        if provided, then consider the inputs periodic with the given period

    Returns
    -------
    y_smooth : array_like
        smoothed y values at each time t or t_out
    """
    t_input = t
    prep = _prep_smooth(t, y, dy, span, t_out, span_out, period)
    t, y, dy, span, t_out, span_out, indices = prep
    if period:
        t_input = np.asarray(t_input) % period

    w = dy ** -2
    w, yw, tw, tyw, ttw = windowed_sum([w, y * w, w, y * w, w], t=t,
                                       tpowers=[0, 0, 1, 1, 2],
                                       span=span, indices=indices,
                                       subtract_mid=cv, period=period)

    denominator = (w * ttw - tw * tw)
    slope = (tyw * w - tw * yw)
    intercept = (ttw * yw - tyw * tw)

    if t_out is None:
        return (slope * t_input + intercept) / denominator
    elif span_out is not None:
        return (slope * t_out + intercept) / denominator
    else:
        i = np.minimum(len(t) - 1, np.searchsorted(t, t_out))
        return (slope[i] * t_out + intercept[i]) / denominator[i]


def multinterp(x, y, xquery, slow=False):
    """Multiple linear interpolations

    Parameters
    ----------
    x : array_like, shape=(N,)
        sorted array of x values
    y : array_like, shape=(N, M)
        array of y values corresponding to each x value
    xquery : array_like, shape=(M,)
        array of query values
    slow : boolean, default=False
        if True, use slow method (used mainly for unit testing)

    Returns
    -------
    yquery : ndarray, shape=(M,)
        The interpolated values corresponding to each x query.
    """
    x, y, xquery = map(np.asarray, (x, y, xquery))
    assert x.ndim == 1
    assert xquery.ndim == 1
    assert y.shape == x.shape + xquery.shape

    # make sure xmin < xquery < xmax in all cases
    xquery = np.clip(xquery, x.min(), x.max())

    if slow:
        from scipy.interpolate import interp1d
        return np.array([interp1d(x, y)(xq) for xq, y in zip(xquery, y.T)])
    elif len(x) == 3:
        # Most common case: use a faster approach
        yq_lower = y[0] + (xquery - x[0]) * (y[1] - y[0]) / (x[1] - x[0])
        yq_upper = y[1] + (xquery - x[1]) * (y[2] - y[1]) / (x[2] - x[1])
        return np.where(xquery < x[1], yq_lower, yq_upper)
    else:
        i = np.clip(np.searchsorted(x, xquery, side='right') - 1,
                    0, len(x) - 2)
        j = np.arange(len(xquery))
        return y[i, j] + ((xquery - x[i]) *
                          (y[i + 1, j] - y[i, j]) / (x[i + 1] - x[i]))





"""Python implementation of Friedman's Supersmoother"""
from __future__ import absolute_import

__version__ = '0.3'

from .smoother import MovingAverageSmoother, LinearSmoother
from .supersmoother import SuperSmoother



from __future__ import division, print_function, absolute_import
import numpy as np
from .utils import linear_smooth, moving_average_smooth, iterable

__all__ = ['MovingAverageSmoother', 'LinearSmoother']


class Smoother(object):
    """Base Class for Smoothers"""
    def __init__(self):
        raise NotImplementedError()

    def fit(self, t, y, dy=1, presorted=False):
        """Fit the smoother

        Parameters
        ----------
        t : array_like
            time locations of the points to smooth
        y : array_like
            y locations of the points to smooth
        dy : array_like or float (default = 1)
            Errors in the y values
        presorted : bool (default = False)
            If True, then t is assumed to be sorted.

        Returns
        -------
        self : Smoother instance
        """
        self.t, self.y, self.dy = self._validate_inputs(t, y, dy, presorted)
        self._fit(self.t, self.y, self.dy)
        return self

    def predict(self, t):
        """Predict the smoothed function value at time t

        Parameters
        ----------
        t : array_like
            Times at which to predict the result

        Returns
        -------
        y : ndarray
            Smoothed values at time t
        """
        t = np.asarray(t)
        return self._predict(np.ravel(t)).reshape(t.shape)

    def cv_values(self, cv=True):
        """Return the values of the cross-validation for the fit data"""
        return self._cv_values(cv)

    def cv_residuals(self, cv=True):
        """Return the residuals of the cross-validation for the fit data"""
        vals = self.cv_values(cv)
        return (self.y - vals) / self.dy

    def cv_error(self, cv=True, skip_endpoints=True):
        """Return the sum of cross-validation residuals for the input data"""
        resids = self.cv_residuals(cv)
        if skip_endpoints:
            resids = resids[1:-1]
        return np.mean(abs(resids))

    def _validate_inputs(self, t, y, dy, presorted=False):
        t, y, dy = np.broadcast_arrays(t, y, dy)
        if presorted:
            self.isort = slice(None)
        elif hasattr(self, 'period') and self.period:
            self.isort = np.argsort(t % self.period)
        else:
            self.isort = np.argsort(t)
        return t[self.isort], y[self.isort], dy[self.isort]

    def _fit(self, t, y, dy):
        """Private function to perform fit() on input data"""
        raise NotImplementedError()

    def _predict(self, t):
        """Private function implementing prediction for new data"""
        raise NotImplementedError()

    def _cv_values(self, cv=True):
        """Private function implementing cross-validation on fit data"""
        raise NotImplementedError()


class SpannedSmoother(Smoother):
    """Base class for smoothers based on local spans of sorted data"""
    def __init__(self, span, period=None):
        self.span = span
        self.period = period

    def _fit(self, t, y, dy):
        pass

    def span_int(self, t=None):
        if t is None:
            t = self.t

        if callable(self.span):
            spanint = self.span(t) * len(self.t)
        elif iterable(self.span):
            spanint = self.span[self.isort] * len(self.t)
        else:
            spanint = self.span * len(self.t)

        return np.clip(spanint, 3, None)

    def _predict(self, t):
        if callable(self.span):
            return self._smoothfunc(self.t, self.y, self.dy, cv=False,
                                    span_out=self.span_int(t), t_out=t,
                                    period=self.period)
        else:
            return self._smoothfunc(self.t, self.y, self.dy, cv=False,
                                    span=self.span_int(), t_out=t,
                                    period=self.period)

    def _cv_values(self, cv=True):
        return self._smoothfunc(self.t, self.y, self.dy, cv=cv,
                                span=self.span_int(), period=self.period)


class MovingAverageSmoother(SpannedSmoother):
    """Local smoother based on a moving average of adjacent points

    Parameters
    ----------
    span : float, array, or function
        The fraction of the data to use at each point of the smooth.
        If a function is passed, then this will be evaluated at each input
        time to determine the smooth.
    period : float (optional)
        If specified, then use a periodic smoother with the given period.
        Default is to assume no periodicity.
    """
    _smoothfunc = staticmethod(moving_average_smooth)


class LinearSmoother(SpannedSmoother):
    """Local smoother based on a locally linear fit of adjacent points

    Parameters
    ----------
    span : float, array, or function
        The fraction of the data to use at each point of the smooth.
        If a function is passed, then this will be evaluated at each input
        time to determine the smooth.
    period : float (optional)
        If specified, then use a periodic smoother with the given period.
        Default is to assume no periodicity.
    """
    _smoothfunc = staticmethod(linear_smooth)



import numpy as np
import matplotlib.pyplot as plt
from supersmoother import MovingAverageSmoother, LinearSmoother, SuperSmoother

rng = np.random.RandomState(0)
N = 10
period = 1
t = np.linspace(0, period, N, endpoint=False)
y = rng.rand(N)
y = np.arange(N)

t_folded = np.concatenate([-period + t, t, t + period])
y_folded = np.concatenate([y, y, y])

model1 = SuperSmoother(primary_spans=[0.05/3, 0.2/3, 0.5/3],
                       middle_span=0.2/3, final_span=0.5/3, period=None)
model2 = SuperSmoother(period=period)

model1.fit(t_folded, y_folded)
model2.fit(t, y)

print(model1.span_int())
print(model2.span_int())

plt.plot(t_folded, y_folded, '.k')
#plt.plot(model1.t, model1.cv_values())
#plt.plot(model2.t, model2.cv_values())

plt.plot(t, model1.predict(t))
plt.plot(t, model2.predict(t))

plt.show()
exit()

rng = np.random.RandomState(0)
N = 10
period = 1
t = np.linspace(0, period, N, endpoint=False)
y = rng.rand(N)

t_folded = np.concatenate([-period + t[N / 2:], t, t[:N / 2] + period])
y_folded = np.concatenate([y[N / 2:], y, y[:N / 2]])


fig, ax = plt.subplots(3, 2, figsize=(8, 8))

for j, Model in enumerate([MovingAverageSmoother, LinearSmoother]):
    for i, span in enumerate([3, 4, 5]):
        model1 = Model(span / len(t_folded), period=None)
        model2 = Model(span / len(t), period=period)
        ax[i, j].plot(t, model1.fit(t_folded, y_folded).predict(t))
        ax[i, j].plot(t, model2.fit(t, y). predict(t))

plt.show()



from __future__ import division, absolute_import
from .. import utils
from ..utils import windowed_sum, windowed_sum_slow

import numpy as np
from numpy.testing import assert_allclose, assert_, assert_equal, assert_raises


def test_iterable():
    x = [1, 2, 3]
    y = 1
    assert_(utils.iterable(x))
    assert_(not utils.iterable(y))


def test_multiinterp(rseed=0, N=100):
    """Test of the multinterp function"""
    rng = np.random.RandomState(rseed)

    def check_match(k):
        x = np.sort(rng.rand(k))
        y = rng.rand(k, N)
        xquery = rng.rand(N)

        res1 = utils.multinterp(x, y, xquery, slow=False)
        res2 = utils.multinterp(x, y, xquery, slow=True)
        assert_allclose(res1, res2)

    for k in 3, 4, 5:
        yield check_match, k


def test_setattr_context():
    """Test of the setattr_context() function"""
    class Foo(object):
        pass
    f = Foo()
    f.attr = "abc"
    with utils.setattr_context(f, attr="123"):
        assert_equal(f.attr, "123")
    assert_equal(f.attr, "abc")

    try:
        with utils.setattr_context(f, attr="123"):
            raise ValueError()
    except:
        pass
    assert_equal(f.attr, "abc")


def test_validate_inputs_nosort(N=10, rseed=0):
    """Test input validation without sorting"""
    rng = np.random.RandomState(rseed)
    t, y = rng.rand(2, N)
    dy = 1

    t_, y_, dy_ = utils.validate_inputs(t, y, dy)
    assert_allclose(t, t_)
    assert_allclose(y, y_)
    assert_allclose(dy, dy_)


def test_validate_inputs_sort(N=10, rseed=0):
    """Test input validation with sorting"""
    rng = np.random.RandomState(rseed)
    t, y = rng.rand(2, N)
    dy = 1

    t_, y_, dy_ = utils.validate_inputs(t, y, dy, sort_by=t)
    isort = np.argsort(t)
    assert_allclose(t[isort], t_)
    assert_allclose(y[isort], y_)
    assert_allclose(dy, dy_)


def test_validate_inputs_fail():
    rng = np.random.RandomState(0)
    t, y = rng.rand(2, 10)
    dy = 1

    # bad keyword argument
    assert_raises(ValueError, utils.validate_inputs, t, y, dy,
                  blah='blah')

    # non-sortable array
    assert_raises(ValueError, utils.validate_inputs, 1, 1, 1)

    # bad sort array
    assert_raises(ValueError, utils.validate_inputs, t, y, dy, sort_by=[1])


def _fixed_span_sum(a, span, subtract_mid=False):
    """fixed-span sum; used for testing windowed_span"""
    window = np.ones(span)
    if span <= len(a):
        res = np.convolve(a, window, mode='same')
    else:
        start = (span - 1) // 2
        res = np.convolve(a, window, mode='full')[start:start + len(a)]
    if subtract_mid:
        res -= a
    return res


def _variable_span_sum(a, span, subtract_mid=False):
    """variable-span sum, using multiple runs of _fixed_span_sum"""
    a, spans = np.broadcast_arrays(a, span)
    unique_spans, inv = np.unique(spans, return_inverse=True)
    results = [_fixed_span_sum(a, s, subtract_mid) for s in unique_spans]
    return np.asarray([results[j][i] for i, j in enumerate(inv)])


def test_windowed_sum_simple():
    #"""windowed sum in the simplest case"""
    rng = np.random.RandomState(0)
        
    def check_results(wsum, N, span):
        a = rng.rand(N)
        assert_allclose([_fixed_span_sum(a, span),
                         _fixed_span_sum(a ** 2, span)],
                        wsum([a, a ** 2], span))

    for wsum in [windowed_sum, windowed_sum_slow]:
        for N in range(2, 6):
            for span in range(1, 7):
                yield check_results, wsum, N, span


def test_windowed_sum_with_t():
    #"""windowed sum with powers of t"""
    rng = np.random.RandomState(0)
        
    def check_results(wsum, N, span):
        a = rng.rand(N)
        t = np.sort(rng.rand(N))
        assert_allclose([_fixed_span_sum(a, span),
                         _fixed_span_sum(a * t, span)],
                        wsum([a, a], span, t=t, tpowers=[0, 1]))

    for wsum in [windowed_sum, windowed_sum_slow]:
        for N in range(2, 6):
            for span in range(1, 7):
                yield check_results, wsum, N, span


def test_windowed_sum_varspan():
    #"""windowed sum with a variable span"""
    rng = np.random.RandomState(0)
        
    def check_results(wsum, N):
        a = rng.rand(N)
        span = rng.randint(1, 5, N)
        assert_allclose([_variable_span_sum(a, span),
                         _variable_span_sum(a ** 2, span)],
                        wsum([a, a ** 2], span))

    for wsum in [windowed_sum, windowed_sum_slow]:
        for N in range(2, 6):
            yield check_results, wsum, N


def test_windowed_sum_with_period():
    #"""windowed sum with period"""
    rng = np.random.RandomState(0)
        
    def check_results(wsum, N, span, tpowers, period=0.6):
        a = rng.rand(N)
        t = np.sort(period * rng.rand(N))

        # find the periodic result straightforwardly by concatenating arrays
        a_all = np.concatenate([a for i in range(6)])
        t_all = np.concatenate([t + i * period for i in range(-3, 3)])

        res1 = wsum(len(tpowers) * [a_all], span, t=t_all,
                    tpowers=tpowers, period=None)
        res1 = [res[3 * N: 4 * N] for res in res1]
        res2 = wsum(len(tpowers) * [a], span, t=t,
                    tpowers=tpowers, period=period)
        assert_allclose(res1, res2)

    for wsum in [windowed_sum, windowed_sum_slow]:
        for N in range(4, 7):
            for span in range(1, 7):
                for tpowers in [(0, 1), (0, 1, 2)]:
                    yield check_results, wsum, N, span, tpowers
    

def test_windowed_sum_with_indices():
    #"""windowed sum with indices"""
    rng = np.random.RandomState(0)
        
    def check_results(wsum, N, span):
        ind = rng.randint(0, N, 4)
        a = rng.rand(N)
        t = np.sort(rng.rand(N))

        res1 = wsum([a, a], span, t=t, tpowers=[0, 1])
        res1 = [res[ind] for res in res1]
        res2 = wsum([a, a], span, t=t, tpowers=[0, 1],
                                 indices=ind)
        
        assert_allclose(res1, res2)

    for wsum in [windowed_sum, windowed_sum_slow]:
        for N in range(4, 6):
            for span in range(1, 7):
                yield check_results, wsum, N, span


def test_windowed_sum_subtract_mid():
    #"""windowed sum, subtracting the middle item"""
    rng = np.random.RandomState(0)
        
    def check_results(wsum, N, span):
        a = rng.rand(N)
        t = np.sort(rng.rand(N))
        assert_allclose([_fixed_span_sum(a, span, subtract_mid=True),
                         _fixed_span_sum(a * t, span, subtract_mid=True)],
                        wsum([a, a], span, t=t, tpowers=[0, 1],
                             subtract_mid=True))

    for wsum in [windowed_sum, windowed_sum_slow]:
        for N in range(2, 6):
            for span in range(1, 7):
                yield check_results, wsum, N, span


def test_windowed_sum_fast_vs_slow():
    #"""Test fast vs slow windowed sum for many combinations of parameters"""
    rng = np.random.RandomState(0)
        
    def check_results(N, span, indices, subtract_mid,
                      tpowers, period, use_t):
        a = rng.rand(N)
        t = np.sort(rng.rand(N))
        if period:
            t *= period

        if np.asarray(tpowers).size == 1:
            arrs = [a]
        else:
            arrs = len(tpowers) * [a]

        if not use_t:
            t = None

        res1 = windowed_sum_slow(arrs, span, t=t, subtract_mid=subtract_mid,
                                 indices=indices, tpowers=tpowers,
                                 period=period)
        res2 = windowed_sum(arrs, span, t=t, subtract_mid=subtract_mid,
                            indices=indices, tpowers=tpowers, period=period)
        assert_allclose(res1, res2)

    for N in [3, 5, 7]:
        for span in [3, 5, rng.randint(2, 5, N)]:
            for indices in [None, rng.randint(0, N, N)]:
                for subtract_mid in [True, False]:
                    for tpowers in [0, (0, 1)]:
                        for period in (0.6, None):
                            for use_t in [True, False]:
                                if period and not use_t:
                                    continue
                                yield (check_results, N, span, indices,
                                       subtract_mid, tpowers, period, use_t)


def test_windowed_sum_bad_args(N=10):
    #"""windowed sum in the simplest case"""
    rng = np.random.RandomState(0)
    a = rng.rand(N)
    t = np.arange(N)

    for wsum in [windowed_sum, windowed_sum_slow]:
        wsum(arrays=[a, a], span=5, t=t)
        assert_raises(ValueError, wsum, arrays=[a, a], span=5, t=t[:-1])
        assert_raises(ValueError, wsum, arrays=[a, a[:-1]], span=5, t=t)

    for wsum in [windowed_sum, windowed_sum_slow]:
        wsum(arrays=[a, a], span=5, tpowers=[0, 1])
        assert_raises(ValueError, wsum, arrays=[a, a],
                      span=5, tpowers=[0, 1, 2])


def make_linear(N=100, err=1E-6, rseed=None):
    rng = np.random.RandomState(rseed)
    t = 10 * rng.rand(N)
    y = t + err * rng.randn(N)
    return t, y, err


def test_perfect_sine(N=100, rseed=0):
    """Test approximate recovery of a sine wave with no noise"""
    t = np.linspace(0, 10, 200)
    y = np.sin(t)
    dy = 1

    for method in [utils.moving_average_smooth, utils.linear_smooth]:
        yfit = method(t, y, dy, span=5)
        assert_allclose(y[3:-3], yfit[3:-3], atol=0.005)


def test_constant_data(N=100, rseed=0):
    """Test linear & moving average smoothers on constant data"""
    rng = np.random.RandomState(rseed)
    t = 10 * rng.rand(N)
    y = 1.0
    dy = 1.0
    span = 5

    for method in [utils.moving_average_smooth, utils.linear_smooth]:
        yfit = method(t, y, dy, span=5)
        assert_allclose(y, yfit)


def test_equal_spaced_linear_data(N=100, rseed=0):
    """Test linear & moving average smoothers on equally-spaced linear data"""
    t = np.linspace(0, 10, N)
    y = 1 + 2 * t
    dy = 1
    span = 5

    for method in [utils.moving_average_smooth, utils.linear_smooth]:
        yfit = method(t, y, dy, span=5)
        assert_allclose(y[3:-3], yfit[3:-3])


def test_random_linear_data(N=100, rseed=0):
    """Test linear smoother on linear data"""
    rng = np.random.RandomState(rseed)
    t = 10 * rng.rand(N)
    y = 1 + 2 * t
    dy = 1.0
    span = 5

    yfit = utils.linear_smooth(t, y, dy, span=5)
    assert_allclose(y, yfit)


def test_t_vs_t_out(N=100, rseed=0):
    """Test two means of arriving at the same answer"""
    rng = np.random.RandomState(rseed)
    t = np.sort(10 * rng.rand(N))
    y = np.sin(t)
    dy = 1.0
    span = 5

    for method in [utils.linear_smooth, utils.moving_average_smooth]:
        yfit1 = method(t, y, dy, span=5, cv=False)
        yfit2 = method(t, y, dy, span_out=5, t_out=t, cv=False)
        assert_allclose(yfit1, yfit2)


def test_smooth_assertions():
    """Test errors and assertions in linear_smooth and moving_average_smooth"""
    rng = np.random.RandomState(0)
    t = 10 * rng.rand(10)
    y = np.sin(t)
    dy = 1.0

    for method in [utils.linear_smooth, utils.moving_average_smooth]:
        # No span set
        assert_raises(ValueError, method, t, y, dy)

        # Both spans set
        assert_raises(ValueError, method, t, y, dy, span=1,
                      t_out=t, span_out=1)

        # span_out set without t_out
        assert_raises(ValueError, method, t, y, dy, span_out=1)



from distutils.core import setup

DESCRIPTION = "Python implementation of Friedman's Supersmoother"
LONG_DESCRIPTION = """
SuperSmoother in Python
=======================
This is an efficient implementation of Friedman's SuperSmoother based in
Python. It makes use of numpy for fast numerical computation.

For more information, see the github project page:
http://github.com/jakevdp/supersmoother
"""
NAME = "supersmoother"
AUTHOR = "Jake VanderPlas"
AUTHOR_EMAIL = "jakevdp@uw.edu"
MAINTAINER = "Jake VanderPlas"
MAINTAINER_EMAIL = "jakevdp@uw.edu"
URL = 'http://github.com/jakevdp/supersmoother'
DOWNLOAD_URL = 'http://github.com/jakevdp/supersmoother'
LICENSE = 'BSD 3-clause'

import supersmoother
VERSION = supersmoother.__version__

setup(name=NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      maintainer=MAINTAINER,
      maintainer_email=MAINTAINER_EMAIL,
      url=URL,
      download_url=DOWNLOAD_URL,
      license=LICENSE,
      packages=['supersmoother',
                'supersmoother.tests',
            ],
      classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4'],
     )



from __future__ import absolute_import, division

import numpy as np
from numpy.testing import assert_allclose, assert_array_less, assert_equal

from .. import SuperSmoother, LinearSmoother


def make_sine(N=100, err=0.05, rseed=None):
    rng = np.random.RandomState(rseed)
    t = 2 * np.pi * rng.rand(N)
    y = np.sin(t) + err * rng.randn(N)
    return t, y, err


def make_linear(N=100, err=1E-6, rseed=None):
    rng = np.random.RandomState(rseed)
    t = 10 * rng.rand(N)
    y = t + err * rng.randn(N)
    return t, y, err


def test_sine():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)

    tfit = np.linspace(1, 5.3, 50)
    ytrue = np.sin(tfit)

    model = SuperSmoother().fit(t, y, dy)
    yfit = model.predict(tfit)
    obs_err = np.mean((yfit - ytrue) ** 2)
    assert_array_less(obs_err, 0.001)


def test_sine_cv():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)
    ytrue = np.sin(np.sort(t)[1:-1])

    model = SuperSmoother().fit(t, y, dy)
    yfit = model.cv_values()[1:-1]
    obs_err = np.mean((yfit - ytrue) ** 2)
    assert_array_less(obs_err, 0.001)


def test_line_linear():
    t, y, dy = make_linear(err=1E-6, rseed=0)
    tfit = np.linspace(0, 10, 40)

    model = SuperSmoother().fit(t, y, dy)
    yfit = model.predict(tfit)
    assert_allclose(tfit, yfit, atol=1E-5)


def test_bass_enhancement_zero():
    """Test low-level bass enhancement"""
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)

    model1 = SuperSmoother(alpha=0).fit(t, y, dy)
    model2 = SuperSmoother().fit(t, y, dy)

    assert_allclose(model1.predict(t), model2.predict(t), atol=1E-3)


def test_bass_enhancement_10():
    """Test high-level bass enhancement"""
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)

    model1 = SuperSmoother(alpha=10).fit(t, y, dy)
    model2 = LinearSmoother(span=0.5).fit(t, y, dy)

    assert_allclose(model1.predict(t), model2.predict(t), atol=1E-1)



def test_cv_interface():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)
    model = SuperSmoother().fit(t, y, dy)

    def check_model(cv, skip_endpoints):
        cv_values = model.cv_values(cv=cv)
        cv_residuals = (model.y - cv_values) / model.dy
        if skip_endpoints:
            cv_error = np.mean(abs(cv_residuals[1:-1]))
        else:
            cv_error = np.mean(abs(cv_residuals))

        assert_allclose(cv_residuals,
                        model.cv_residuals(cv=cv))
        assert_allclose(cv_error,
                        model.cv_error(cv=cv, skip_endpoints=skip_endpoints))

    for cv in [True, False]:
        for skip_endpoints in [True, False]:
            yield check_model, cv, skip_endpoints



from __future__ import division, print_function, absolute_import

import numpy as np

from .smoother import LinearSmoother
from .utils import multinterp, setattr_context


class SuperSmoother(LinearSmoother):
    """SuperSmoother for nonparametric smoothing of scatter plots.

    SuperSmoother is an adaptive component-wise linear smoother, described
    in Friedman 1984 [1].

    Parameters
    ----------
    alpha : float (optional)
        If specified, the bass enhancement / smoothing level (0 < alpha < 10).
    period : float (optional)
        If specified, then assume the data is periodic with the given period.

    Other Parameters
    ----------------
    primary_spans : array_like, default=(0.05, 0.2, 0.5)
        The primary span values used for the smooth. Must be between 0 and 1.
        Only modify these if you know exactly what you're doing!
    middle_span : float (default = 0.2)
        The middle span value used for the algorithm.
        Only modify this if you know exactly what you're doing!
    final_span : float (default = 0.05)
        The final span value used in the algorithm.
        Only modify this if you know exactly what you're doing!

    Attributes
    ----------
    primary_smooths : list of LinearSmoother
        the trained LinearSmoother instances used for the initial data pass
    span : function
        a function giving the optimal span at a given t value
        (available only after fit() is called)

    [1] Friedman 1984, "A Variable Span Smoother"
        http://www.slac.stanford.edu/cgi-wrap/getdoc/slac-pub-3477.pdf
    """
    def __init__(self, alpha=None, period=None,
                 primary_spans=(0.05, 0.2, 0.5),
                 middle_span=0.2, final_span=0.05):
        self.alpha = alpha
        self.period = period
        self.primary_spans = np.sort(np.unique(primary_spans))
        self.middle_span = middle_span
        self.final_span = final_span
        self.primary_smooths = [LinearSmoother(span, period)
                                for span in self.primary_spans]

    def _fit(self, t, y, dy):
        mid_smooth = LinearSmoother(self.middle_span, self.period)

        # 1. Compute three smoothed curves and get residuals
        ysmooth_primary = np.array([smoother.fit(t, y, dy, presorted=True)
                                            .cv_values()
                                    for smoother in self.primary_smooths])
        resids = abs((ysmooth_primary - y) / dy)

        # 2. Smooth each set of residuals with the midrange
        smoothed_resids = np.array([mid_smooth.fit(t, resid, dy,
                                                   presorted=True)
                                              .cv_values(cv=False)
                                    for resid in resids])

        # 3. Select span yielding best residual at each point
        best_spans = self.primary_spans[np.argmin(smoothed_resids, 0)]

        # 3a. Apply bass enhancement, if necessary
        if self.alpha is not None:
            alpha = np.clip(self.alpha, 0, 10)
            minresid = smoothed_resids.min(0)
            factor = (minresid / smoothed_resids[-1]) ** (10 - alpha)
            best_spans += factor * (self.primary_spans[-1] - best_spans)

        # 4. Smooth best span estimates with midrange span
        self.spansmoother = LinearSmoother(self.middle_span, self.period)
        self.spansmoother.fit(t, best_spans, dy, presorted=True)
        smoothed_spans = self.spansmoother.cv_values(cv=False)

        # 5. Interpolate the smoothed values
        smoothed_spans = np.clip(smoothed_spans,
                                 self.primary_spans[0],
                                 self.primary_spans[-1])
        self.ysmooth_raw = multinterp(self.primary_spans,
                                      ysmooth_primary,
                                      smoothed_spans)

        # The final smooth is done over self.ysmooth_raw with a constant span
        # width given by self.final_span.
        # We'll make this happen in _predict() and _cv_values() below.

        # for convenience in accessing the span values (and to make subclass
        # functions behave) set span to the spansmoother predict function.
        self.span = self.spansmoother.predict

    def _predict(self, t):
        # temporarily change y and span:
        with setattr_context(self, y=self.ysmooth_raw, span=self.final_span):
            return LinearSmoother._predict(self, t)

    def _cv_values(self, cv=True):
        # temporarily change y and span:
        with setattr_context(self, y=self.ysmooth_raw, span=self.final_span):
            # Use cv=False in all circumstances: CV is built into ysmooth_raw.
            return LinearSmoother._cv_values(self, cv=False)



from __future__ import absolute_import, division

import numpy as np
from numpy.testing import (assert_allclose, assert_array_less,
                           assert_equal, assert_raises)

from .. import MovingAverageSmoother, LinearSmoother
from ..smoother import Smoother


def make_sine(N=100, err=0.05, rseed=None):
    rng = np.random.RandomState(rseed)
    t = 2 * np.pi * rng.rand(N)
    y = np.sin(t) + err * rng.randn(N)
    return t, y, err


def make_linear(N=100, err=1E-6, rseed=None):
    rng = np.random.RandomState(rseed)
    t = 10 * rng.rand(N)
    y = t + err * rng.randn(N)
    return t, y, err


def test_sine():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)

    tfit = np.linspace(1, 5.3, 50)
    ytrue = np.sin(tfit)

    def check_model(Model, span, err):
        model = Model(span).fit(t, y, dy)
        yfit = model.predict(tfit)
        obs_err = np.mean((yfit - ytrue) ** 2)
        assert_array_less(obs_err, err)

    spans = [0.05, 0.2, 0.5]
    errs = [0.005, 0.01, 0.1]

    for Model in [MovingAverageSmoother, LinearSmoother]:
        for span, err in zip(spans, errs):
            yield check_model, Model, span, err


def test_sine_cv():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)
    ytrue = np.sin(np.sort(t)[1:-1])

    def check_model(Model, span, err):
        model = Model(span).fit(t, y, dy)
        yfit = model.cv_values()[1:-1]
        obs_err = np.mean((yfit - ytrue) ** 2)
        assert_array_less(obs_err, err)

    spans = [0.05, 0.2, 0.5]
    errs = [0.005, 0.02, 0.1]

    for Model in [MovingAverageSmoother, LinearSmoother]:
        for span, err in zip(spans, errs):
            yield check_model, Model, span, err


def test_line_linear():
    t, y, dy = make_linear(rseed=0, err=1E-6)
    tfit = np.linspace(0, 10, 40)

    def check_model(span):
        model = LinearSmoother(span).fit(t, y, dy)
        yfit = model.predict(tfit)
        assert_allclose(tfit, yfit, atol=1E-5)

    for span in [0.05, 0.2, 0.5]:
        yield check_model, span



def test_variable_sine_predict():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)

    tfit = np.linspace(1, 5.3, 50)
    ytrue = np.sin(tfit)

    def check_model(Model, span):
        model1 = Model(span * np.ones_like(t))
        model2 = Model(span)
        assert_allclose(model1.fit(t, y, dy).predict(tfit),
                        model2.fit(t, y, dy).predict(tfit))

    for Model in [MovingAverageSmoother, LinearSmoother]:
        for span in [0.05, 0.2, 0.5]:
            yield check_model, Model, span


def test_variable_sine_crossval():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)

    tfit = np.linspace(1, 5.3, 50)
    ytrue = np.sin(tfit)

    def check_model(Model, span):
        model1 = Model(span * np.ones_like(t))
        model2 = Model(span)
        assert_allclose(model1.fit(t, y, dy).cv_values(),
                        model2.fit(t, y, dy).cv_values())

    for Model in [MovingAverageSmoother, LinearSmoother]:
        for span in [0.05, 0.2, 0.5]:
            yield check_model, Model, span


def test_func_span_const():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)

    def check_model(Model, span, spanfunc):
        model1 = Model(span)
        model2 = Model(spanfunc)
        assert_allclose(model1.fit(t, y, dy).predict(t),
                        model2.fit(t, y, dy).predict(t))

    for Model in [MovingAverageSmoother, LinearSmoother]:
        for span in [0.05, 0.2, 0.5]:
            spanfunc = lambda t, span=span: span * np.ones_like(t)
            yield check_model, Model, span, spanfunc


def test_func_span_variable():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)

    spanfunc = lambda t, tmax=t.max(): 0.5 + 0.05 * t / tmax
    span = spanfunc(t)

    def check_model(Model, span, spanfunc):
        model1 = Model(span)
        model2 = Model(spanfunc)
        assert_allclose(model1.fit(t, y, dy).predict(t),
                        model2.fit(t, y, dy).predict(t))

    for Model in [MovingAverageSmoother, LinearSmoother]:
        yield check_model, Model, span, spanfunc


def test_cv_interface():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)
    span = 0.2

    def check_model(Model, span, cv, skip_endpoints):
        model = Model(span).fit(t, y, dy)
        cv_values = model.cv_values(cv=cv)
        cv_residuals = (model.y - cv_values) / model.dy
        if skip_endpoints:
            cv_error = np.mean(abs(cv_residuals[1:-1]))
        else:
            cv_error = np.mean(abs(cv_residuals))

        assert_allclose(cv_residuals,
                        model.cv_residuals(cv=cv))
        assert_allclose(cv_error,
                        model.cv_error(cv=cv, skip_endpoints=skip_endpoints))

    for Model in [MovingAverageSmoother, LinearSmoother]:
        for cv in [True, False]:
            for skip_endpoints in [True, False]:
                yield check_model, Model, span, cv, skip_endpoints


def test_span_extremes():
    """Test models with extreme values of spans"""
    t, y, dy = make_sine(10, err=0.05, rseed=0)

    def check_model(Model, span):
        model1 = Model(span).fit(t, y, dy)
        model2 = Model(span + np.zeros_like(t)).fit(t, y, dy)
        assert_allclose(model1.cv_values(False), model2.cv_values(False))

    for Model in [MovingAverageSmoother, LinearSmoother]:
        for span in (-1, 0, 1, 2):
            yield check_model, Model, span
    

def test_periodic():
    rng = np.random.RandomState(0)
    N = 10
    period = 1
    t = np.linspace(0, period, N, endpoint=False)
    y = rng.rand(N)

    t_folded = np.concatenate([-period + t, t, t + period])
    y_folded = np.concatenate([y, y, y])

    def check_model(Model, span):
        model1 = Model(span / len(t_folded), period=None)
        model2 = Model(span / len(t), period=period)

        model1.fit(t_folded, y_folded)
        model2.fit(t, y)

        assert_allclose(model1.predict(t), model2.predict(t))
    

    for Model in [MovingAverageSmoother, LinearSmoother]:
        for span in [3, 4, 5]:
            yield check_model, Model, span


def test_baseclass():
    # silly tests to bring coverage to 100%
    assert_raises(NotImplementedError, Smoother)
    class Derived(Smoother):
        def __init__(self):
            pass

    d = Derived()
    assert_raises(NotImplementedError, d.fit, [1, 2, 3], [1, 1, 1])
    assert_raises(NotImplementedError, d.predict, [1, 2, 3])
    assert_raises(NotImplementedError, d.cv_values)
